from django.apps import AppConfig
from django.db.models.signals import post_migrate, pre_migrate


class MpathyConfig(AppConfig):
    name = 'mpathy'

    def ready(self):
        from .operations import inject_pre_migration_operations, inject_post_migration_operations

        pre_migrate.connect(inject_pre_migration_operations, sender=self)
        post_migrate.connect(inject_post_migration_operations, sender=self)



from django.contrib.postgres.indexes import GistIndex
from django.db import models
from django.db.models.expressions import CombinedExpression, RawSQL

from .fields import LTree, LTreeField, Subpath


class BadMove(ValueError):
    pass


class MPathQuerySet(models.QuerySet):
    def get_cached_trees(self):
        """
        Evaluates this queryset and returns a list of top-level nodes.
        """
        nodes_by_path = {}
        min_level = None
        for node in self:
            node._cached_children = []
            nodes_by_path[node.ltree] = node
            level = node.ltree.level()
            if min_level is None or min_level > level:
                min_level = level

        top_level_nodes = []
        for node in self:
            parent_path = node.ltree.parent()
            if parent_path in nodes_by_path:
                nodes_by_path[parent_path]._cached_children.append(node)
            elif node.ltree.level() == min_level:
                top_level_nodes.append(node)

        return top_level_nodes


class MPathManager(models.Manager.from_queryset(MPathQuerySet)):
    def move_subtree(self, node, new_parent):
        """
        Moves a node and all its descendants under the given new parent.
        If the parent is None, the node will become a root node.

        If the `node` is equal to the new parent, or is an ancestor of it,
        raises BadMove.

        If node's parent is already new_parent, returns immediately.

        NOTE:
        This updates all the nodes in the database, and the current node instance.
        It cannot update any other node instances that are in memory, so if you have some
        whose ltree paths are affected by this function you may need to refresh them
        from the database.
        """
        if node.is_ancestor_of(new_parent, include_self=True):
            raise BadMove(
                "%r can't be made a child of %r"
                % (node.ltree, new_parent.ltree if new_parent else None)
            )

        # Check if there's actually anything to do, return if not
        if node.parent_id is None:
            if new_parent is None:
                return
        elif new_parent is not None and node.ltree.parent() == new_parent.ltree:
            return

        old_parent_ltree = node.ltree.parent()
        old_parent_level = old_parent_ltree.level() if old_parent_ltree else -1

        # An expression which refers to the part of a given node's path which is
        # not in the current node's old parent path.
        # i.e. when node is 'a.b.c', the old parent will be 'a.b',
        # so for a descendant called 'a.b.c.d.e' we want to find 'c.d.e'.
        ltree_tail_expr = Subpath(models.F("ltree"), old_parent_level + 1)

        # Update the ltree on all descendant nodes to match new_parent
        qs = self.filter(ltree__descendant_or_equal=node.ltree)
        if new_parent is None:
            new_ltree_expr = ltree_tail_expr
        else:
            # TODO: how to do this without raw sql? django needs a cast expression
            # here otherwise the concat() fails because the ltree is interpreted
            # as text. Additionally, there's no available concatenation operator for ltrees
            # exposable in django.
            new_ltree_expr = CombinedExpression(
                lhs=RawSQL("%s::ltree", [new_parent.ltree]),
                connector="||",
                rhs=ltree_tail_expr,
            )
        qs.update(
            ltree=new_ltree_expr,
            # Update parent at the same time as ltree, otherwise the check constraint fails
            parent=models.Case(
                models.When(
                    pk=node.pk,
                    then=models.Value(new_parent.ltree if new_parent else None),
                ),
                default=Subpath(new_ltree_expr, 0, -1),
                output_field=node.__class__._meta.get_field("parent"),
            ),
        )

        # Update node in memory
        node.parent = new_parent
        node._set_ltree()
        node.save(update_fields=["parent"])


class MPathNode(models.Model):
    ltree = LTreeField(null=False, unique=True)
    label = models.CharField(null=False, blank=False, max_length=255)

    # Duplicating the whole path is annoying, but we need this field so that the
    # database can ensure consistency when we create a node.
    # Otherwise, we could create 'a.b' without first creating 'a'.
    parent = models.ForeignKey(
        "self",
        related_name="children",
        null=True,
        to_field="ltree",
        db_index=False,
        on_delete=models.CASCADE,
    )

    objects = MPathManager()

    class Meta:
        abstract = True
        indexes = [
            GistIndex(fields=["ltree"]),
            GistIndex(fields=["parent"]),
        ]

    def _set_ltree(self):
        if self.parent_id:
            ltree = "%s.%s" % (self.parent_id, self.label)
        else:
            ltree = self.label
        self.ltree = LTree(ltree)

    def save(self, **kwargs):
        # If no label, let the db throw an error. Otherwise, ensure path is consistent with
        # parent and label.
        if not self.label:
            raise ValueError(
                "%s objects must have a label. Got: label=%r"
                % (self.__class__.__name__, self.label)
            )

        # If this is a new node or parent has changed, re-calculate ltree.
        # NOTE: This only works for leaf nodes, so shouldnt be relied on.
        # It's here as a convenience so you can create nodes with a pre-set parent.
        # For *changing* parent of an existing node, use MPathManager.move_subtree()
        self._set_ltree()

        return super(MPathNode, self).save(**kwargs)

    def is_ancestor_of(self, other, include_self=False):
        if other is None:
            return False
        return self.ltree.is_ancestor_of(other.ltree, include_self=include_self)

    def is_descendant_of(self, other, include_self=False):
        if other is None:
            return True
        return self.ltree.is_descendant_of(other.ltree, include_self=include_self)

    def get_siblings(self, include_self=False):
        """
        Returns a queryset of this node's siblings, using the default manager.

        If include_self=True is given, the queryset will include this node.
        """
        mgr = self.__class__._default_manager
        qs = mgr.filter(parent__ltree=self.parent_id)

        if not include_self:
            qs = qs.exclude(ltree=self.ltree)
        return qs

    def get_children(self):
        """
        Returns a queryset of children for this node, using the default manager.
        """
        try:
            # Shortcut the database if this node has been fetched using
            # qs.get_cached_trees()
            return self._cached_children
        except AttributeError:
            mgr = self.__class__._default_manager
            return mgr.filter(ltree__lquery=self.ltree.children_lquery())

    def get_descendants(self, include_self=False):
        """
        Returns a queryset of descendants for this node, using the default manager.

        If include_self=True is given, the queryset will include this node.
        """
        mgr = self.__class__._default_manager
        qs = mgr.filter(ltree__descendant_or_equal=self.ltree)

        if not include_self:
            qs = qs.exclude(ltree=self.ltree)
        return qs

    def get_ancestors(self, include_self=False):
        """
        Returns a queryset of ancestors for this node, using the default manager.

        If include_self=True is given, the queryset will include this node.
        """
        mgr = self.__class__._default_manager
        qs = mgr.filter(ltree__ancestor_or_equal=self.ltree)

        if not include_self:
            qs = qs.exclude(ltree=self.ltree)
        return qs



from psycopg2.extensions import quote_ident

from django.apps import apps as global_apps
from django.contrib.postgres.operations import CreateExtension
from django.core.exceptions import FieldDoesNotExist
from django.db import connection, DEFAULT_DB_ALIAS, migrations

from .fields import LTreeField
from .models import MPathNode


class LTreeExtension(CreateExtension):

    def __init__(self):
        self.name = 'ltree'


def inject_pre_migration_operations(plan=None, apps=global_apps, using=DEFAULT_DB_ALIAS, **kwargs):
    """
    Insert a `LTreeExtension` operation before every planned `CreateModel` operation.
    """
    if plan is None:
        return

    for migration, backward in plan:
        for index, operation in enumerate(migration.operations):
            if isinstance(operation, migrations.CreateModel):
                for name, field in operation.fields:
                    if isinstance(field, LTreeField):
                        migration.operations.insert(index, LTreeExtension())
                        return


def post_migrate_mpathnode(model):
    # Note: model *isn't* a subclass of MPathNode, because django migrations are Weird.
    # if not issubclass(model, MPathNode):
    # Hence the following workaround:
    try:
        ltree_field = model._meta.get_field('ltree')
        if not isinstance(ltree_field, LTreeField):
            return
    except FieldDoesNotExist:
        return

    names = {
        "table": quote_ident(model._meta.db_table, connection.connection),
        "check_constraint": quote_ident('%s__check_ltree' % model._meta.db_table, connection.connection),
    }

    cur = connection.cursor()
    # Check that the ltree is always consistent with being a child of _parent
    cur.execute('''
        ALTER TABLE %(table)s ADD CONSTRAINT %(check_constraint)s CHECK (
            (parent_id IS NOT NULL AND ltree ~ (parent_id::text || '.*{1}')::lquery)
            OR (parent_id IS NULL AND ltree ~ '*{1}'::lquery)
        )
    ''' % names)


def inject_post_migration_operations(plan=None, apps=global_apps, using=DEFAULT_DB_ALIAS, **kwargs):
    if plan is None:
        return

    for migration, backward in plan:
        for index, operation in reversed(list(enumerate(migration.operations))):
            if isinstance(operation, migrations.CreateModel):
                model = apps.get_model(migration.app_label, operation.name)
                post_migrate_mpathnode(model)



#!/usr/bin/env python
from setuptools import setup
import os
import re


def get_version():
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join("mpathy", "__init__.py")).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


setup(
    name="django-mpathy",
    description="""Simple hierarchical data using materialised path, in a postgres database.""",
    version=get_version(),
    author="Craig de Stigter",
    author_email="craig.ds@gmail.com",
    url="http://github.com/craigds/django-mpathy",
    packages=["mpathy"],
    install_requires=[
        "psycopg2>=2.7",
        "Django>=2.2",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Database",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Utilities",
    ],
)



from django.db import models


class Subpath(models.Func):
    function = "subpath"


class LTree(str):
    def labels(self):
        return self.split(".")

    def level(self):
        """
        Returns the level of this node.
        Root nodes are level 0.
        """
        return self.count(".")

    def is_root(self):
        return self.level() == 0

    def is_ancestor_of(self, other, include_self=False):
        if include_self and self == other:
            return True
        if other.level() <= self.level():
            return False
        return other.labels()[: self.level() + 1] == self.labels()

    def is_descendant_of(self, other, include_self=False):
        return other.is_ancestor_of(self, include_self=include_self)

    def parent(self):
        if self.level():
            return LTree(self.rsplit(".", 1)[0])
        return None

    def children_lquery(self):
        """
        Returns an lquery which finds nodes which are children of the current node.
        """
        return "%s.*{1}" % self

    def parent_lquery(self):
        """
        Returns an lquery to find the parent of the current node.
        """
        return ".".join(self.labels()[:-1])


class LTreeField(models.CharField):
    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = 256
        super(LTreeField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(LTreeField, self).deconstruct()
        del kwargs["max_length"]
        return name, path, args, kwargs

    def db_type(self, connection):
        return "ltree"

    def to_python(self, value):
        return LTree(value)

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return LTree(value)


class Level(models.Transform):
    lookup_name = "level"
    function = "nlevel"

    @property
    def output_field(self):
        return models.IntegerField()


class LQuery(models.Lookup):
    lookup_name = "lquery"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s ~ %s" % (lhs, rhs), params


class DescendantOrEqual(models.Lookup):
    lookup_name = "descendant_or_equal"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s <@ %s" % (lhs, rhs), params


class AncestorOrEqual(models.Lookup):
    lookup_name = "ancestor_or_equal"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s @> %s" % (lhs, rhs), params


LTreeField.register_lookup(Level)
LTreeField.register_lookup(LQuery)
LTreeField.register_lookup(DescendantOrEqual)
LTreeField.register_lookup(AncestorOrEqual)
